/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 F. JEZEQUEL, J.-L. LAMOTTE, J.E. TIDSWELL
Copyright 2023 F. JEZEQUEL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
#ifndef __SAM__
#define __SAM__

//#include <algorithm>
#include <type_traits>
#include <iostream>
#include <stack>
#include <stdio.h>
#include <cmath>
#include <time.h>
#include <stdint.h>
#include <string.h>
#include <math.h>

#include "gmp.h"
#include "mpfr.h"

#include "SAM_private.h"

#define SAM_DIV       0x1
#define SAM_MUL       0x2
#define SAM_POWER     0x4
#define SAM_MATH      0x8
#define SAM_BRANCHING 0x10
#define SAM_INTRINSIC 0x20
#define SAM_CANCEL    0x40
#define SAM_ALL       0xFFFFFFFF



using namespace std;



////////////////////

template<int  T>     // forward de la classe template Optional<T>
class mp_st;


//MP_ST_NBBITS_MAX returns the maximum of 2 precisions S and T
//MP_ST_NBBITS_MAX is used for instance to compare 2 mp_st variables
template <int S, int T> struct  MP_ST_NBBITS_MAX
{
    const static  int      value = ((S > T)  ? S : T);
};


// a short cut way to declare type of max bits
// typename MP_ST_MAX<S,T>::type
template <int S, int T> struct  MP_ST_MAX
{
    typedef mp_st< MP_ST_NBBITS_MAX<S,T>::value > type;
};


//
//  Simplify Template expansion by identifying special case versions 
//

template <typename T>
struct SAM_IS_DETERMINISTIC { const static bool value = std::is_arithmetic< typename std::decay<T>::type >::value; };
template <>
struct SAM_IS_DETERMINISTIC <mpfr_srcptr> { const static bool value = true; };
template <>
struct SAM_IS_DETERMINISTIC <mpfr_ptr> { const static bool value = true; };
template<>
struct SAM_IS_DETERMINISTIC <mpfr_t> { const static bool value = true; };


template <typename T>
struct SAM_IS_STOCHASTIC { const static bool value = false; };
template<int S>
struct SAM_IS_STOCHASTIC < mp_st<S> > { const static bool value = true; };




template<int T>
class  mp_st
{
    template<int> friend class mp_st;
    
protected:
    mpfr_t x,y,z;    
    mutable int accuracy;
    
    
public:
    
    // Constructors
    // ASSIGNMENT
    
    
#include "SAM_to.h"
    
    
    inline ~mp_st() 
    {
        if ( (x)->_mpfr_d != 0 ) { // move semantics might zero this
            mpfr_clears(x, y, z, (mpfr_ptr)0 );
            (x)->_mpfr_d = 0;
        }
    } 
    
    inline mpfr_t* getX()  { return &x;};
    inline mpfr_t* getY()  { return &y;};
    inline mpfr_t* getZ()  { return &z;};
    inline mpfr_prec_t  getPrecision() const { return T;};
    inline int getAccuracy(){ return accuracy;  }
    
#include "SAM_add.h"
#include "SAM_sub.h"
#include "SAM_mul.h"
#include "SAM_div.h"
    
#include "SAM_eq.h"
#include "SAM_ne.h"
#include "SAM_ge.h"
#include "SAM_gt.h"
#include "SAM_le.h"
#include "SAM_lt.h"
#include "SAM_convert.h"
#include "SAM_digitnumber.h"
#include "SAM_numericalnoise.h"    
#include "SAM_str.h"
    
#include "SAM_intr.h"
#include "SAM_math.h"
#include "SAM_fma.h"
    
};


void  SAM_init(int, unsigned int, unsigned int, unsigned int);
void  SAM_init(int, unsigned int);
void  SAM_init(int);
void  SAM_end();

void self_validation_only();
void SAM_enable(unsigned int);
void SAM_disable(unsigned int);

template <int  T>
std::ostream& operator<<(std::ostream&, const mp_st<T>&);

template <int  T>
std::istream& operator >>(std::istream&, mp_st<T>& );



//****m* SAM_str/operator<<
//    NAME
//     operator<<
//
//    SYNOPSIS
//      ostream& operator <<(ostream&, const mp_st &) 
//    FUNCTION
//      << operator for stochastic variables
//    INPUTS
//   
//    RESULT
//      void
//    SEE ALSO
//      str(3)
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//      
//      
//  /



#include "SAM_op2.h"

#include "SAM_str2.h"



#endif

