/*-----------------------------------------------------------------------
Copyright 2020 J Tidswell

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// Provides simple template wrapper for common MPFR functions to make code appear more generic
//
///////////////////////////////////////////////////

#ifndef __SAM_MPFR_TEMPLATES__
#define __SAM_MPFR_TEMPLATES__

#include "mpfr.h"
#include <cassert>

// use a private namespace so that the MPFR definitions specifically for SAM dont clash with other projects
namespace SAM {
        
    /* stack allocate small mpfr */
    /* modified version of the official macro */
    /*
     * #define MPFR_DECL_INIT(_x, _p)					\
     *		MPFR_EXTENSION mp_limb_t __gmpfr_local_tab_##_x[((_p)-1)/GMP_NUMB_BITS+1]; \
     *		MPFR_EXTENSION mpfr_t _x = {{(_p),1,__MPFR_EXP_NAN,__gmpfr_local_tab_##_x}}
     */
    inline void mpfr_init_local(mpfr_t &x, mp_limb_t* limb )
    {
	const unsigned short float_signficand_size = 24;
	assert(GMP_NUMB_BITS>float_signficand_size); /*default is 32 of 64 so this is reasonable */
	x[0]._mpfr_prec = float_signficand_size;
	x[0]._mpfr_sign = 1;
	x[0]._mpfr_exp = __MPFR_EXP_NAN;             
	x[0]._mpfr_d = limb;
    }
    inline void mpfr_clear_local(mpfr_t &x) 
    {
	// useful for error catching, disable in release builds ?
	x->_mpfr_d = NULL;
	x->_mpfr_exp = __MPFR_EXP_NAN;
	x->_mpfr_prec = 0;
	x->_mpfr_sign = 0;
    }
    
    
    /* borrowing more from MPFR */
    // MPFR does not have a kind function for general init variables ...
    inline mpfr_kind_t mpfr_get_kind (const mpfr_t& x)
    {
	const mpfr_exp_t k = x->_mpfr_exp;
	mpfr_kind_t kind;
	if ( k == __MPFR_EXP_NAN ) { kind = MPFR_NAN_KIND; }
	else if ( k == __MPFR_EXP_INF ) { kind = MPFR_INF_KIND; }
	else if ( k == __MPFR_EXP_ZERO ) { kind = MPFR_ZERO_KIND; }
	else /* k > __MPFR_EXP_INF */ { kind = MPFR_REGULAR_KIND; }
	return kind;
    }
    
    
    
    //
    //  Addition
    //
    inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_add_ui( res, src, a, dirn ); }
    inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) { return mpfr_add_si( res, src, a, dirn ); }
    inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_add_ui( res, src, a, dirn ); }
    inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) { return mpfr_add_si( res, src, a, dirn ); }
    inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_add_d( res, src, a, dirn ); }
    inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_add_d( res, src, (double)a, dirn ); }
    
    inline int add_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_t& a, mpfr_rnd_t dirn ) { return mpfr_add( res, src, a, dirn ); }
    
    //
    //  Subtraction
    //
    
    inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_sub_ui( res, src, a, dirn ); }
    inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) {    return mpfr_sub_si( res, src, a, dirn );}
    inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_sub_ui( res, src, a, dirn ); }
    inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) {    return mpfr_sub_si( res, src, a, dirn );}
    inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_sub_d( res, src, a, dirn ); }
    inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_sub_d( res, src, (double)a, dirn ); }

    inline int sub_mpfr( mpfr_ptr res, const unsigned int& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_ui_sub( res, a, src, dirn ); }
    inline int sub_mpfr( mpfr_ptr res, const int& a, mpfr_srcptr src, mpfr_rnd_t dirn ) {    return mpfr_si_sub( res, a, src, dirn );}
    inline int sub_mpfr( mpfr_ptr res, const unsigned long& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_ui_sub( res, a, src, dirn ); }
    inline int sub_mpfr( mpfr_ptr res, const long& a, mpfr_srcptr src, mpfr_rnd_t dirn ) {    return mpfr_si_sub( res, a, src, dirn );}
    inline int sub_mpfr( mpfr_ptr res, const double& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_d_sub( res, a, src, dirn ); }
    inline int sub_mpfr( mpfr_ptr res, const float& a, mpfr_srcptr src, mpfr_rnd_t dirn ) { return mpfr_d_sub( res, (double)a, src, dirn ); }
    
    inline int sub_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_t& a, mpfr_rnd_t dirn ) { return mpfr_sub( res, src, a, dirn ); }

    //
    //  Multiplication
    //

    inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_mul_ui( res, src, a, dirn ); }
    inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) { return mpfr_mul_si( res, src, a, dirn ); }
    inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_mul_ui( res, src, a, dirn ); }
    inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) { return mpfr_mul_si( res, src, a, dirn ); }
    inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_mul_d( res, src, a, dirn ); }
    inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_mul_d( res, src, (double)a, dirn ); }
    
    inline int mul_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_mul( res, src, a, dirn ); }
    
    
    //
    //  Division
    //
    
    inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_div_ui( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const int& a, mpfr_rnd_t dirn ) { return mpfr_div_si( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_div_ui( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const long& a, mpfr_rnd_t dirn ) { return mpfr_div_si( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const double& a, mpfr_rnd_t dirn ) { return mpfr_div_d( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const float& a, mpfr_rnd_t dirn ) { return mpfr_div_d( res, src, (double)a, dirn ); }

    inline int div_mpfr( mpfr_ptr res, const unsigned int& src, mpfr_srcptr a, mpfr_rnd_t dirn ) { return mpfr_ui_div( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, const int& src, mpfr_srcptr a, mpfr_rnd_t dirn ) { return mpfr_si_div( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, const unsigned long& src, mpfr_srcptr a, mpfr_rnd_t dirn ) { return mpfr_ui_div( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, const long& src, mpfr_srcptr a, mpfr_rnd_t dirn ) { return mpfr_si_div( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, const double& src, mpfr_srcptr a, mpfr_rnd_t dirn ) { return mpfr_d_div( res, src, a, dirn ); }
    inline int div_mpfr( mpfr_ptr res, const float& src, mpfr_srcptr a, mpfr_rnd_t dirn ) { return mpfr_d_div( res, (double)src, a, dirn ); }

    inline int div_mpfr( mpfr_ptr res, mpfr_srcptr src, const mpfr_srcptr& a, mpfr_rnd_t dirn ) { return mpfr_div( res, src, a, dirn ); }


    //
    //  Construction
    //
    
    inline int set_mpfr( mpfr_ptr x, const unsigned int& a, mpfr_rnd_t dirn ) { return mpfr_set_ui( x, a, dirn ); }
    inline int set_mpfr( mpfr_ptr x, const int& a, mpfr_rnd_t dirn ) { return mpfr_set_si( x, a, dirn ); }
    inline int set_mpfr( mpfr_ptr x, const unsigned long& a, mpfr_rnd_t dirn ) { return mpfr_set_ui( x, a, dirn ); }
    inline int set_mpfr( mpfr_ptr x, const long& a, mpfr_rnd_t dirn ) { return mpfr_set_si( x, a, dirn ); }
    inline int set_mpfr( mpfr_ptr x, const double& a, mpfr_rnd_t dirn ) { return mpfr_set_d( x, a, dirn ); }
    inline int set_mpfr( mpfr_ptr x, const float& a, mpfr_rnd_t dirn ) { return mpfr_set_d( x, (double)a, dirn ); }
    
    inline int set_mpfr( mpfr_ptr x, const mpfr_t& a, mpfr_rnd_t dirn ) { return mpfr_set( x, a, dirn ); }
    
    
    //
    //  power
    //
    
    template<typename T>
    inline int pow_mpfr( mpfr_ptr res, mpfr_srcptr b, const T& a, mpfr_rnd_t dirn );
    
    template<> inline int pow_mpfr<unsigned int>( mpfr_ptr res, mpfr_srcptr a, const unsigned int& b, mpfr_rnd_t dirn ) { return mpfr_pow_ui( res, a, b, dirn ); }
    template<> inline int pow_mpfr<int>( mpfr_ptr res, mpfr_srcptr a, const int& b, mpfr_rnd_t dirn ) { return mpfr_pow_si( res, a, b, dirn ); }
    template<> inline int pow_mpfr<unsigned long>( mpfr_ptr res, mpfr_srcptr a, const unsigned long& b, mpfr_rnd_t dirn ) { return mpfr_pow_ui( res, a, b, dirn ); }
    template<> inline int pow_mpfr<long>( mpfr_ptr res, mpfr_srcptr a, const long& b, mpfr_rnd_t dirn ) { return mpfr_pow_si( res, a, b, dirn ); }
    
    template<> inline int pow_mpfr<mpfr_t>( mpfr_ptr res, mpfr_srcptr a, const mpfr_t& b, mpfr_rnd_t dirn ) { return mpfr_pow( res, a, b, dirn ); }
    template<> inline int pow_mpfr<mpfr_srcptr>( mpfr_ptr res, mpfr_srcptr a, const mpfr_srcptr& b, mpfr_rnd_t dirn ) { return mpfr_pow( res, a, b, dirn ); }
    
    //
    // template trick to resolve default precision (digits)
    //
    
    template<typename T> struct mp_st_nbbit {
	// default is unknown ...
	static int prec(const T&) { return -1; } 
    };

    template<> struct mp_st_nbbit<int> {
	static int prec(const int&) { return sizeof(int)*8 -1; } 
    };
    
    template<> struct mp_st_nbbit<unsigned int> {
	static int prec(const unsigned int&) { return sizeof(unsigned int)*8 ; }   
    };
    
    template<> struct mp_st_nbbit<float> {
	static int prec(const float&) { return 24; } 
    };
    template<> struct mp_st_nbbit<double> {
	static int prec(const double&) { return 53; }
    };
    template<> struct mp_st_nbbit<mpfr_t> {
	static int prec(const mpfr_t& u) { return (int)mpfr_get_prec(u); } 
    };
    template<> struct mp_st_nbbit<mpfr_srcptr> {
	static int prec(const mpfr_srcptr& u) { return (int)mpfr_get_prec(u); } 
    };
    
    
    // template <int S> class mp_st;
    
    // template<int S> struct  mp_st_nbbit<mp_st<S>>  {
    // 	static int prec(const mp_st<S>&) { return S; }
    // };
    
    
} // SAM namespace



#endif // __SAM_MPFR_TEMPLATES__
