/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 F. JEZEQUEL, J.-L. LAMOTTE, J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
// 
///////////////////////////////////////////////////
// Last Modifications
// 
//
///////////////////////////////////////////////////


//****f* SAM_digitnumber/nb_significant_digit
//    NAME
//      nb_significant_digit
//    SYNOPSIS
//      res = nb_significant_digit(x) 
//      res = x.nb_significant_digit() 
//    FUNCTION
//      The digitnumber() function returns the number of significant
//      digits of a stochastic x
//      
//      
//    INPUTS
//      x           - a stochastic number
//    RESULT
//      res         - an integer value
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//    SEE ALSO
//      
//      
//  /

int nb_significant_digit() const
{
  // saves the rounding mode to compute the number of exact significant digits
  // always with the same rounding mode:
  mp_rnd_t tmp_rnd= mpfr_get_default_rounding_mode();
  mpfr_set_default_rounding_mode(MPFR_RNDN);
  mp_rnd_t default_rnd =mpfr_get_default_rounding_mode();
  const long MAX_PREC = (long)(((double)T*LOG));

  //  printf("precision %d MAX_PREC =  %ld\n", T, MAX_PREC);
  
  mpfr_t x0, x1, x2, xx;
  mpfr_init2 (x0, T);
  mpfr_init2 (x1, T);
  mpfr_init2 (x2, T);
  mpfr_init2 (xx, T);
  
  mpfr_add (xx, x, y, default_rnd);
  mpfr_add (xx, xx, z, default_rnd);  
  
  accuracy=0;
  if (mpfr_zero_p (xx)!=0){
    if ( (mpfr_equal_p(x,y)!=0) &&(mpfr_equal_p(x,z)!=0) ) {
      accuracy=MAX_PREC; 
    }
  }
  else {
    mpfr_ui_div (xx, 3, xx, default_rnd); 
    mpfr_mul (x0, x, xx, default_rnd);
    mpfr_sub_ui (x0, x0, 1, default_rnd); 
    mpfr_mul (x1, y, xx, default_rnd);
    mpfr_sub_ui (x1, x1, 1, default_rnd); 
    mpfr_mul (x2, z, xx, default_rnd);
    mpfr_sub_ui (x2, x2, 1, default_rnd); 
    
    mpfr_sqr (x0, x0, default_rnd);
    mpfr_sqr (x1, x1, default_rnd);
    mpfr_sqr (x2, x2, default_rnd);
    // reuse xx for yy for speed
    mpfr_add (xx, x0, x1, default_rnd);
    mpfr_add (xx, xx, x2, default_rnd); 
    
    if (mpfr_number_p(xx)==0) {
      accuracy=0;
    }
    else {
      mpfr_mul_d (xx, xx, 3.0854661704166664, default_rnd);
      long e = (long)mpfr_get_exp(xx);

      //      printf("long e:: %ld\n",e);
      if ( e <= -2*(long)T ) {
	accuracy=MAX_PREC; 
	//	printf("MAX acc%d\n",accuracy);
      }
      else {
	if (e<=0)
	  //as proposed in EBFJ15 (Reliable Computing, 21:35–52, 2015):
	  //	  accuracy= (int)(-e * LOG * 0.5);
	  //more pessimistic approximation proposed in May 2020:
	  accuracy= (int)(-(e+2) * LOG * 0.5);
	//	  printf("Comp acc%d\n",accuracy);
      }
    }
  }
  
  mpfr_set_default_rounding_mode(tmp_rnd);
  
  mpfr_clears(x0, x1, x2, xx, (mpfr_ptr) 0);
  
  return accuracy; 
} 


