/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 F. JEZEQUEL, J.-L. LAMOTTE, J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
// 
///////////////////////////////////////////////////
// Last Modification 
//
//
///////////////////////////////////////////////////
//****m* SAM_add/operator+
//    NAME
//      operator+
//    SYNOPSIS
//      res = a + b 
//    FUNCTION
//    Defines all the functions involving at least one
//    stochastic argument which overload the  "+" operator 
//    in a statement such as "a+b" or "+a".      
//      
//      
//    INPUTS
//      a           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      b           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      At least one argument must be of stochastic type.
//     
//    RESULT  
//      res         - a stochastic number
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//    SEE ALSO
//      
//      
//  /


// ADDITION

// with one non-stochastic operand
template<int T, typename argT>
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type
inline operator+( const mp_st<T>& a, const argT& b)
{
    mp_st<T> res(a);
    res += b;
    return res;
}
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline & operator+( mp_st<T>&& a, const argT& b) 
{
    a += b; 
    return a;
}
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline operator+( const argT& a, const mp_st<T>& b ) 
{
    mp_st<T> res(b);
    res += a;
    return res;
}
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline & operator+( const argT& a,  mp_st<T>&& b ) 
{
    b += a;
    return b;
}

// move semantics when both arguments of same size
template<int T> inline mp_st<T>& operator+( mp_st<T>&& a, mp_st<T>&& b ) { a += b; return a; }
template<int T> inline mp_st<T>& operator+( mp_st<T>&& a, const mp_st<T>& b ) { a += b; return a; }
template<int T> inline mp_st<T>& operator+( const mp_st<T>& a, mp_st<T>&& b ) { b += a; return b; }
template<int T> inline mp_st<T> operator+( const mp_st<T>& a, const mp_st<T>& b ) { mp_st<T> res(a); res += b; return res; }

// move semantics when larger is moveable
template <int S, int T>
typename std::enable_if< (S>T), mp_st<S> >::type
inline & operator+(mp_st<S> && a, const mp_st<T>& b)
{
    a += b;
    return a;
}
template <int S, int T>
typename std::enable_if< (S<T), mp_st<T> >::type
inline & operator+(const mp_st<S> & a, mp_st<T>&& b)
{
    b += a;
    return b;
}
// when sizes are different, and the larger is not moveable, need a temporary
template <int S, int T>
typename std::enable_if< (S!=T), typename MP_ST_MAX<S,T>::type >::type
inline operator+(const mp_st<S> & a, const mp_st<T>& b)
{
    // the operation should be done in the larger precision
    typename MP_ST_MAX<S,T>::type res(a);
    res += b;
    return res;
}


// SUBTRACTION

// with one non-stochastic operand
template<int T, typename argT>
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type
inline operator-( const mp_st<T>& a, const argT& b) { mp_st<T> res(a); res -= b; return res; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline & operator-( mp_st<T>&& a, const argT& b) { a -= b; return a; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline operator-( const argT& a, const mp_st<T>& b ) { mp_st<T> res(b); res -= a; return -res; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline & operator-( const argT& a,  mp_st<T>&& b ) { b -= a; b.negate(); return b; }

// move semantics when both arguments of same size
template<int T> inline mp_st<T>& operator-( mp_st<T>&& a, mp_st<T>&& b ) { a -= b; return a; }
template<int T> inline mp_st<T>& operator-( mp_st<T>&& a, const mp_st<T>& b ) { a -= b; return a; }
template<int T> inline mp_st<T>& operator-( const mp_st<T>& a, mp_st<T>&& b ) { b -= a; b.negate(); return b; }
template<int T> inline mp_st<T> operator-( const mp_st<T>& a, const mp_st<T>& b ) { mp_st<T> res(a); res -= b; return res; }

// move semantics when larger is moveable
template <int S, int T>
typename std::enable_if< (S>T), mp_st<S> >::type
inline & operator-(mp_st<S> && a, const mp_st<T>& b)
{
    a -= b;
    return a;
}
template <int S, int T>
typename std::enable_if< (S<T), mp_st<T> >::type
inline & operator-(const mp_st<S> & a, mp_st<T>&& b)
{
    b -= a;
    return -b;
}
// when sizes are different, and the larger is not moveable, need a temporary
template <int S, int T>
typename std::enable_if< (S!=T), typename MP_ST_MAX<S,T>::type >::type
inline operator-(const mp_st<S> & a, const mp_st<T>& b)
{
    // the operation should be done in the larger precision
    typename MP_ST_MAX<S,T>::type res(a);
    res -= b;
    return res;
}


// MULTIPLICATION

// with one non-stochastic operand
template<int T, typename argT>
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type
inline operator*( const mp_st<T>& a, const argT& b) { mp_st<T> res(a); res *= b; return res; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline & operator*( mp_st<T>&& a, const argT& b) { a *= b; return a; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline operator*( const argT& a, const mp_st<T>& b ) { mp_st<T> res(b); res *= a; return res; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline & operator*( const argT& a,  mp_st<T>&& b ) { b *= a; return b; }

// move semantics when both arguments of same size
template<int T> inline mp_st<T>& operator*( mp_st<T>&& a, mp_st<T>&& b ) { a *= b; return a; }
template<int T> inline mp_st<T>& operator*( mp_st<T>&& a, const mp_st<T>& b ) { a *= b; return a; }
template<int T> inline mp_st<T>& operator*( const mp_st<T>& a, mp_st<T>&& b ) { b *= a; return b; }
template<int T> inline mp_st<T> operator*( const mp_st<T>& a, const mp_st<T>& b ) { mp_st<T> res(a); res *= b; return res; }

// move semantics when larger is moveable
template <int S, int T>
typename std::enable_if< (S>T), mp_st<S> >::type
inline & operator*(mp_st<S> && a, const mp_st<T>& b)
{
    a *= b;
    return a;
}
template <int S, int T>
typename std::enable_if< (S<T), mp_st<T> >::type
inline & operator*(const mp_st<S> & a, mp_st<T>&& b)
{
    b *= a;
    return b;
}
// when sizes are different, and the larger is not moveable, need a temporary
template <int S, int T>
typename std::enable_if< (S!=T), typename MP_ST_MAX<S,T>::type >::type
inline operator*(const mp_st<S> & a, const mp_st<T>& b)
{
    // the operation should be done in the larger precision
    typename MP_ST_MAX<S,T>::type res(a);
    res *= b;
    return res;
}



// DIVISION

// with one non-stochastic operand
template<int T, typename argT>
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type
inline operator/( const mp_st<T>& a, const argT& b) { mp_st<T> res(a); res /= b; return res; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline & operator/( mp_st<T>&& a, const argT& b) { a /= b; return a; }
template<int T, typename argT> 
typename std::enable_if< SAM_IS_DETERMINISTIC<argT>::value, mp_st<T> >::type 
inline operator/( const argT& a, const mp_st<T>& b ) { mp_st<T> res(a); res /= b; return res; }


// move semantics when both arguments of same size
template<int T> inline mp_st<T>& operator/( mp_st<T>&& a, mp_st<T>&& b ) { a /= b; return a; }
template<int T> inline mp_st<T>& operator/( mp_st<T>&& a, const mp_st<T>& b ) { a /= b; return a; }
template<int T> inline mp_st<T> operator/( const mp_st<T>& a, const mp_st<T>& b ) { mp_st<T> res(a); res /= b; return res; }

// move semantics when larger is moveable
template <int S, int T>
typename std::enable_if< (S>T), mp_st<S> >::type
inline & operator/(mp_st<S> && a, const mp_st<T>& b)
{
    a /= b;
    return a;
}
template <int S, int T>
typename std::enable_if< (S<T), mp_st<T> >::type
inline & operator/(const mp_st<S> & a, mp_st<T>&& b)
{
    // the operation should be done in the larger precision
    mp_st<T> res(a);
    res /= b;
    return res;
}
// when sizes are different, and the larger is not moveable, need a temporary
template <int S, int T>
typename std::enable_if< (S!=T), typename MP_ST_MAX<S,T>::type >::type
inline operator/(const mp_st<S> & a, const mp_st<T>& b)
{
    // the operation should be done in the larger precision
    typename MP_ST_MAX<S,T>::type res(a);
    res /= b;
    return res;
}



// Order relations

template<int T, typename argT> inline int operator==(const argT& a, mp_st<T>& b)  { return b.operator==(a); }
template<int T, typename argT> inline int operator!=(const argT& a, mp_st<T>& b)  { return b.operator!=(a); }

// note
// a<=b     equiv b>=a  

template<int T, typename argT> inline int operator<=(const argT& a, mp_st<T>& b)  { return  b.operator>=(a); }
template<int T, typename argT> inline int operator>=(const argT& a, mp_st<T>& b)  { return b.operator<=(a); }
template<int T, typename argT> inline int operator<(const argT& a, mp_st<T>& b)  { return b.operator>(a); }
template<int T, typename argT> inline int operator>(const argT& a, mp_st<T>& b)  { return b.operator<(a); }













