/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 F. JEZEQUEL, J.-L. LAMOTTE, J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
//
///////////////////////////////////////////////////
// Last Modifications
// 
//
///////////////////////////////////////////////////


#include "SAM_private.h"


//////////////////////////////////////////////////////////
//****m* SAM_str/display
//    NAME
//      display
//
//    SYNOPSIS
//      display()
//      display(char *)
//
//    FUNCTION
//      The display method prints the triplet associated with 
//      a stochastic variable.
//    INPUTS
//   
//    RESULT
//      void
//    SEE ALSO
//      str(3)
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//      
//      
//  

void display() const 
{ 
    int default_prec_digit = T *LOG;
    mpfr_out_str(stdout, 10, default_prec_digit+2, x, MPFR_RNDN);
    cout<<" -- ";
    mpfr_out_str(stdout, 10, default_prec_digit+2, y, MPFR_RNDN);
    cout<<" -- ";
    mpfr_out_str(stdout, 10, default_prec_digit+2, z, MPFR_RNDN);
	cout << "\n";
}

///////////////////////////////////////////////////////////////////////////////////////////

void display(const char *s) const 
{ 
    int default_prec_digit = T *LOG;
    cout<< s << "   ";
    mpfr_out_str(stdout, 10, default_prec_digit+2, x, MPFR_RNDN);
    cout<<" -- ";
    mpfr_out_str(stdout, 10, default_prec_digit+2, y, MPFR_RNDN);
    cout<<" -- ";
    mpfr_out_str(stdout, 10, default_prec_digit+2, z, MPFR_RNDN);
}



//////////////////////////////////////////////////////////

//****f* SAM_str/str
//    NAME
//     str
//
//    SYNOPSIS
//      char* str(char *s) const 
//
//      char* str(char *, mp_st&)
//    FUNCTION
//      The output string contains the scientific notation of the 
//      stochastic argument; only the exact significant digits appear 
//      in the string. 
//    INPUTS
//      The str function has a string argument and a stochastic argument.
//    RESULT
//      It returns a pointer to the first argument.
//    SEE ALSO
//      str(3)
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//      
//      
//  /

char* str(char *s) const 
{
	int nres, naux;
	long int nn;
	mpfr_t fract_res,res,sum;
	
	int tmp;
	char *t, tt[__SAM_MAXNBSTRING ];

	mp_rnd_t default_rnd = mpfr_get_default_rounding_mode();
	mpfr_set_default_rounding_mode(MPFR_RNDN);
	mp_rnd_t tmp_rnd = mpfr_get_default_rounding_mode();
	
	mpfr_t x_mpfr,y_mpfr,z_mpfr,im,two,ten,poww;
	mpfr_init2(fract_res, T);
	mpfr_init2(res,T);
	mpfr_init2(sum,T);
	mpfr_init2(x_mpfr,T);
	mpfr_init2(y_mpfr,T);
	mpfr_init2(z_mpfr,T);
	mpfr_init2(im,T);
	mpfr_init2(two,T);
	mpfr_init2(ten,T);
	mpfr_init2(poww,T);
	
	mpfr_set(x_mpfr,x,tmp_rnd);
	mpfr_set(y_mpfr,y,tmp_rnd); 
	mpfr_set(z_mpfr,z,tmp_rnd);

	if (accuracy==DIGIT_NOT_COMPUTED){
	    this->nb_significant_digit();
	}

	strcpy(tt,"                     ");
	if (accuracy==0) {
		strcpy(tt," @.0");
	}
	else {
	    // A VOIR pourquoi cette ligne de code existe. 
	    // accuracy=(accuracy<(int)(T*LOG)) ? accuracy : (int)(T*LOG);

		mpfr_add(sum,x_mpfr,y_mpfr,tmp_rnd);
		mpfr_add(sum,sum,z_mpfr,tmp_rnd);
		mpfr_div_d(res,sum,3.0,tmp_rnd);
		//mpfr_clear(x_mpfr);
		//mpfr_clear(y_mpfr);
		//mpfr_clear(z_mpfr);
		//mpfr_clear(sum);
		
		if (mpfr_cmp_d(res,0.0)<0) strcpy(tt,"-0. ");
		else strcpy(tt," 0. ");
		mpfr_abs(res,res,tmp_rnd);
		
		if (mpfr_cmp_d(res,0.0)==0){
			nn=0;
			mpfr_set_d(fract_res,0.0,tmp_rnd);
		}
		else{
			
			mpfr_log10(im,res,tmp_rnd);
			
			if (mpfr_cmp_d(res,1.0)>=0) nn=mpfr_get_si(im,MPFR_RNDZ)+1;
			else nn=mpfr_get_si(im,MPFR_RNDZ);
			
			mpfr_set_d(ten,10.,tmp_rnd);
			mpfr_pow_si(poww,ten,-nn,tmp_rnd);
			mpfr_mul(fract_res,res,poww,tmp_rnd);

			//mpfr_clear(res);
			//mpfr_clear(pow);
			//mpfr_clear(two);
			//mpfr_clear(im);

			mpfr_t one,point_one;
			mpfr_init2(one,T);
			mpfr_init2(point_one,T);
			mpfr_set_d(one,1,tmp_rnd);
			mpfr_div(point_one,one,ten,tmp_rnd);
			
			if (mpfr_cmp_d(fract_res,0.1)<0) mpfr_set(fract_res,point_one,tmp_rnd);
			if (mpfr_cmp_d(fract_res,1.0)>=0){
				mpfr_set(fract_res,point_one,tmp_rnd);
				nn=nn+1;
			}
			mpfr_clear(one);
			mpfr_clear(point_one);
		}
		naux= accuracy+3;
		t=tt+3;
		for(int i=4;i<naux;i++){
			mpfr_mul(fract_res,fract_res,ten,tmp_rnd);
			nres=mpfr_get_si(fract_res,MPFR_RNDD);
			*t++=48+nres;
			mpfr_sub_si(fract_res,fract_res,nres,tmp_rnd);
		}
		mpfr_mul(fract_res,fract_res,ten,tmp_rnd);
		tmp=mpfr_get_si(fract_res,tmp_rnd);
		nres = tmp < 9 ? tmp : 9 ;
		*t++=48+nres;
		if(nn>0){
			*t++='E';
			*t++='+';
			stack<int> p;
			while (nn!=0){
				p.push(nn%10);
				nn=nn/10;
			}
			int p_size=p.size();
			for (int i=0; i<p_size; i++){
				*t++=48+p.top();
				p.pop();
			}
		}
		else if(nn<0){
			*t++='E';
			*t++='-';
			stack<int> p;
			nn=-nn;
			while (nn!=0){
				p.push(nn%10);
				nn=nn/10;
			}
			int p_size=p.size();
			for (int i=0; i<p_size; i++){
				*t++=48+p.top();
				p.pop();
			}
		}
		*t='\0';
	}
	mpfr_set_default_rounding_mode(default_rnd);
	mpfr_clears(x_mpfr,y_mpfr,z_mpfr,im,two,ten,poww,fract_res,res,sum, (mpfr_ptr) 0);

	//  sprintf(s,"(p=%d) %s", T, tt);
	// sprintf(s,"%s", tt); //deprecated on MAC OS
	snprintf(s,__SAM_MAXNBSTRING , "%s",tt);
	return(s);
}


char* str(char *s, mp_st& a)
{
	return(a.str(s));
}

