/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 F. JEZEQUEL, J.-L. LAMOTTE, J.E. TIDSWELL

  This file is part of SAM.

  SAM is free software: you can redistribute it and/or modify it
  under the terms of the GNU Lesser General Public License as
  published by the Free Software Foundation, either version 3 of the
  License, or (at your option) any later version.

  SAM is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
  Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with SAM. If not, see
  <http://www.gnu.org/licenses/>.
  -----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
//
//
///////////////////////////////////////////////////
// Last Modification 
// 
//
///////////////////////////////////////////////////
//****m* SAM_sub/operator-
//    NAME
//      operator-
//    SYNOPSIS
//      res = a - b 
//    FUNCTION
//      Define all the functions involving at least one argument
//      of stochastic type which overload the "-" operator
//      in a statement such as "a-b".
//    INPUTS
//      a           - an integer, a float, a double, an mpfr_t object or a stochastic number
//      b           - an integer, a float, a double, an mpfr_t object or a stochastic number 
//      At least one argument must be of stochastic type.
//    RESULT
//      res         - a stochastic number
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//    SEE ALSO
//      
//      
//  /


inline mp_st<T>& negate()
{
    mpfr_neg (this->x, this->x, MPFR_RNDU);
    mpfr_neg (this->y, this->y, MPFR_RNDU);
    mpfr_neg (this->z, this->z, MPFR_RNDU);
    return *this;
}


inline mp_st<T> operator-() const & // l-value / copy
{
    mp_st<T> res(*this);
    res.negate();
    return res ;
}


inline mp_st<T> operator-() && // r-value / move
{
    this->negate();
    return std::move(*this);
}

///////////////////////////////////////////////////////////////////////////////////////////



mp_st<T> operator--()
{

    int initAccuracy = this->accuracy;
    if (_SAM_cancel_tag && (initAccuracy==DIGIT_NOT_COMPUTED)) {
	initAccuracy = this->nb_significant_digit();
    }
  
    accuracy=DIGIT_NOT_COMPUTED;
    if (SAM_RANDOM) mpfr_sub_ui (x, x, 1, MPFR_RNDU);
    else mpfr_sub_ui (x, x, 1, MPFR_RNDD);
    if (SAM_RANDOM){
	mpfr_sub_ui (y, y, 1, MPFR_RNDU);
	mpfr_sub_ui (z, z, 1, MPFR_RNDD);
    }
    else {
	mpfr_sub_ui (y, y, 1, MPFR_RNDD);
	mpfr_sub_ui (z, z, 1, MPFR_RNDU);
    }
  
    if (_SAM_cancel_tag){
	if (initAccuracy >=this->nb_significant_digit()+_SAM_cancel_value)
	    instability(&_SAM_cancel_count);
    }
    return *this;
}



///////////////////////////////////////////////////////////////////////////////////////////



mp_st<T> operator--(int)
{
    mp_st initial=*this;
    
    
    accuracy=DIGIT_NOT_COMPUTED;
    if (SAM_RANDOM) mpfr_sub_ui (x, x, 1, MPFR_RNDU);
    else mpfr_sub_ui (x, x, 1, MPFR_RNDD);
    if (SAM_RANDOM){
	mpfr_sub_ui (y, y, 1, MPFR_RNDU);
	mpfr_sub_ui (z, z, 1, MPFR_RNDD);
    }
    else {
	mpfr_sub_ui (y, y, 1, MPFR_RNDD);
	mpfr_sub_ui (z, z, 1, MPFR_RNDU);
    }
    
    if (_SAM_cancel_tag){
	if (initial.accuracy==DIGIT_NOT_COMPUTED) initial.nb_significant_digit();
	if (initial.accuracy >=  this->nb_significant_digit()+_SAM_cancel_value)
	    instability(&_SAM_cancel_count);
    }
    return initial;
}



///////////////////////////////////////////////////////////////////////////////////////////

template<typename arg> 
mp_st<T>& operator-=(const arg& a)
{

  
    int initAccuracy = this->accuracy;
	if (_SAM_cancel_tag && (initAccuracy==DIGIT_NOT_COMPUTED)) {
		initAccuracy = this->nb_significant_digit();
	}
	
	accuracy=DIGIT_NOT_COMPUTED;
	if (SAM_RANDOM) SAM::sub_mpfr (x, x, a, MPFR_RNDU);
	else SAM::sub_mpfr (x, x, a, MPFR_RNDD);
	if (SAM_RANDOM){
		SAM::sub_mpfr (y, y, a, MPFR_RNDU);
		SAM::sub_mpfr (z, z, a, MPFR_RNDD);
	}
	else {
		SAM::sub_mpfr (y, y, a, MPFR_RNDD);
		SAM::sub_mpfr (z, z, a, MPFR_RNDU);
	}
	if (_SAM_cancel_tag){
		if (initAccuracy >=  this->nb_significant_digit()+_SAM_cancel_value)
			instability(&_SAM_cancel_count);
	}
	return *this;
}

///////////////////////////////////////////////////////////////////////////////////////////o
template<int S>
mp_st<T>& operator-=(const mp_st<S>& a)	
{
    int initAccuracy = this->accuracy;
    if (_SAM_cancel_tag && (initAccuracy==DIGIT_NOT_COMPUTED)) {
	initAccuracy = this->nb_significant_digit();
    }
    
    accuracy=DIGIT_NOT_COMPUTED;
    if (SAM_RANDOM) mpfr_sub (x, x, a.x, MPFR_RNDU);
    else mpfr_sub (x, x, a.x, MPFR_RNDD);
    if (SAM_RANDOM){
	mpfr_sub (y, y, a.y, MPFR_RNDU);
	mpfr_sub (z, z, a.z, MPFR_RNDD);
    }
    else {
	mpfr_sub (y, y, a.y, MPFR_RNDD);
	mpfr_sub (z, z, a.z, MPFR_RNDU);
    } 
        
    if (_SAM_cancel_tag){
	if (a.accuracy==DIGIT_NOT_COMPUTED) a.nb_significant_digit();
	if ((initAccuracy < a.accuracy ? initAccuracy : a.accuracy) >= 
	    nb_significant_digit()+_SAM_cancel_value)
	    instability(&_SAM_cancel_count);
    }
    return *this;
}

///////////////////////////////////////////////////////////////////////////////////////////



