/*-----------------------------------------------------------------------
Copyright 2010-2016 S. GRAILLAT, F. JEZEQUEL
Copyright 2010 S. WANG, Y. ZHU
Copyright 2019-2020 F. JEZEQUEL, J.-L. LAMOTTE, J.E. TIDSWELL

This file is part of SAM.

    SAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or (at your option) any later version.

    SAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
    or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
    Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with SAM. If not, see
    <http://www.gnu.org/licenses/>.
-----------------------------------------------------------------------*/
///////////////////////////////////////////////////
//
// SAM SOFTWARE
//
///////////////////////////////////////////////////
// Last Modifications
//
//
///////////////////////////////////////////////////
//****m* SAM_to/operator=
//    NAME
//      operator=
//    SYNOPSIS
//      res = v
//    FUNCTION
//      Define all the functions involving at least one argument
//      of stochastic type which overload the assignment statement "=".
//      
//    INPUTS
//      a           - an integer, a float, a double, an mpfr_t or an mp_st 
//  
//    RESULT
//      res         - an mp_st
//             
//*****
//   You can use this space for remarks that should not be included
//   in the documentation.
//    EXAMPLE
//      
//  
//    NOTES
//  
//  
//    BUGS
//  
//  
//    SEE ALSO
//      
//      
//  /

//
// constructors
//

#include <math.h>
#include <cmath>

#include "SAM_perturbation.h"

mp_st() 
{
    mpfr_init2 (x,(mpfr_prec_t) T); 
    mpfr_init2 (y,(mpfr_prec_t) T); 
    mpfr_init2 (z,(mpfr_prec_t) T); 
#if defined(COUNT_ALLOCATIONS) && COUNT_ALLOCATIONS
	++SAM::alloc_count;
#endif

    mpfr_set_ui (x,0,MPFR_RNDN);
    mpfr_set_ui (y,0,MPFR_RNDN);
    mpfr_set_ui (z,0,MPFR_RNDN);
    accuracy = (int)( T* LOG );
}


mp_st(const mp_st<T>& u) 
{
    // simple copy
    mpfr_init2 (x, (mpfr_prec_t) T);
    mpfr_init2 (y, (mpfr_prec_t) T);
    mpfr_init2 (z, (mpfr_prec_t) T);
#if defined(COUNT_ALLOCATIONS) && COUNT_ALLOCATIONS
	++SAM::alloc_count;
#endif

    mpfr_set (x, u.x, MPFR_RNDN);
    mpfr_set (y, u.y, MPFR_RNDN);
    mpfr_set (z, u.z, MPFR_RNDN);
    accuracy = u.accuracy;
}

mp_st(mp_st<T>&& u) // move
{
    this->x[0] = u.x[0];
    this->y[0] = u.y[0];
    this->z[0] = u.z[0];
#if defined(COUNT_ALLOCATIONS) && COUNT_ALLOCATIONS
	++SAM::move_count;
#endif
    this->accuracy = u.accuracy;
    u.accuracy = DIGIT_NOT_COMPUTED;
    memset(u.x,0,sizeof(mpfr_t));
    memset(u.y,0,sizeof(mpfr_t));
    memset(u.z,0,sizeof(mpfr_t));
}


//----------------------------------------------------------

// a specific constructor is necessary for mp_st<S> because  3 values are necessary
// to represent  a stochastic number (u.x, u.y, u.z)
// For standard types (int, float, double, mpfr_t, ...) only one ! 

template <int S>
mp_st(const mp_st<S>& u) // copy with new precision value 
{
    mpfr_init2 (x, (mpfr_prec_t) T);
    mpfr_init2 (y, (mpfr_prec_t) T);
    mpfr_init2 (z, (mpfr_prec_t) T);
#if defined(COUNT_ALLOCATIONS) && COUNT_ALLOCATIONS
	++SAM::alloc_count;
#endif

    if ( S >= T ){
	if (SAM_RANDOM) mpfr_set (x, u.x, MPFR_RNDU);
	else mpfr_set (x, u.x, MPFR_RNDD);
	
	if ( SAM_RANDOM ) {
	    mpfr_set (y, u.y, MPFR_RNDU);
	    mpfr_set (z, u.z, MPFR_RNDD);
	}
	else
	    {
		mpfr_set (y, u.y, MPFR_RNDD);
		mpfr_set (z, u.z, MPFR_RNDU);
	    }
	accuracy = u.accuracy < T*LOG  ? u.accuracy : T*LOG ; 
	
    }
    else {
	// precision of u < precision of this
	// copy of the fields
	mpfr_set (x, u.x, MPFR_RNDN);
	mpfr_set (y, u.y, MPFR_RNDN);
	mpfr_set (z, u.z, MPFR_RNDN);
	
	accuracy = u.accuracy;
	
	if (mpfr_equal_p(u.x, u.y)!=0 &&  mpfr_equal_p(u.x, u.z)!=0) {
	    // the 3 fields are equal
	    // perturbation of the fields to  avoid an increase of accuracy
	    // the perturbation is on the bit S+1.
	    // the perturbation doesn't change the accuracy value of the stochastic number
	    // if the three fields are different, the accuracy value is embedded in the stochastic
	    // number and no perturbation is required

	    perturbationLastBit(this, u);

	}
    }
} 
    
//----------------------------------------------------------
template <typename argT>
mp_st<T>(const argT& u)
{
    mpfr_init2(x,(mpfr_prec_t)T);
    mpfr_init2(y,(mpfr_prec_t)T);
    mpfr_init2(z,(mpfr_prec_t)T);
#if defined(COUNT_ALLOCATIONS) && COUNT_ALLOCATIONS
	++SAM::alloc_count;
#endif

    mpfr_prec_t p = SAM::mp_st_nbbit<argT>::prec(u);

    if ( p >=  (mpfr_prec_t) T ) {
	
      p = T; 
      if (SAM_RANDOM) SAM::set_mpfr (x, u, MPFR_RNDU);
      else SAM::set_mpfr (x, u, MPFR_RNDD);
      if ( SAM_RANDOM )
	{
	    SAM::set_mpfr (y, u, MPFR_RNDU);
	    SAM::set_mpfr (z, u, MPFR_RNDD);
	}
      else
	{
	    SAM::set_mpfr (y, u, MPFR_RNDD);
	    SAM::set_mpfr (z, u, MPFR_RNDU);
	}
    }
   else {
       // p < T
       // the size of mp_st<T>  is greater than the size of u
       // u can be exactly coded in the mp_st<T>
       SAM::set_mpfr (x, u, MPFR_RNDN);
       SAM::set_mpfr (y, u, MPFR_RNDN);
       SAM::set_mpfr (z, u, MPFR_RNDN);

       // the 3 fields are equal
       // if u is an integer, the stochastic number is exact
       // if u is a floating point number (float, double, mpfr_t) a  perturbation of the fields to  avoid an increase
       // of the accuracy is necessary
       // the perturbation is on the bit S+1.
       // the perturbation doesn't change the accuracy value of the stochastic number

       p=perturbationLastBit(this, u);
   }
       
    accuracy = (int)( p * LOG );
}



mp_st<T>(const double u1, const double u2, const double u3) 
{
    mpfr_init2(x,(mpfr_prec_t)T);
    mpfr_init2(y,(mpfr_prec_t)T);
    mpfr_init2(z,(mpfr_prec_t)T);
#if defined(COUNT_ALLOCATIONS) && COUNT_ALLOCATIONS
	++SAM::alloc_count;
#endif

    // No perturbation to be compatible with CADNA
    mpfr_set_d(x, u1, MPFR_RNDN);
    mpfr_set_d(y, u2, MPFR_RNDN);
    mpfr_set_d(z, u3, MPFR_RNDN);
  
    accuracy = DIGIT_NOT_COMPUTED;
}


friend mp_st<T>& sam_set_str(mp_st<T> &v, const char *s)
{
    // should be rewritten to take into account the size of s, the type of number (int, float, ... ) in s ....
    // 
    mpfr_t y;

    mpfr_init2(y,T);
    mpfr_set_str(y,s,10,MPFR_RNDN);
    
    SAM::set_mpfr(v.x,y,MPFR_RNDN);
    SAM::set_mpfr(v.y,y,MPFR_RNDN);
    SAM::set_mpfr(v.z,y,MPFR_RNDN);
    mpfr_clear(y);

    v.accuracy = T*LOG;
    //    v.data_st();
    
    return v;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// assignment operators
//

template <int S>
mp_st<T>& operator=(const mp_st<S> &u) // copy with new precision value 
{
    if ( S >= T ){
	if (SAM_RANDOM) mpfr_set (x, u.x, MPFR_RNDU);
	else mpfr_set (x, u.x, MPFR_RNDD);
	
	if ( SAM_RANDOM ) {
	    mpfr_set (y, u.y, MPFR_RNDU);
	    mpfr_set (z, u.z, MPFR_RNDD);
	}
	else
	    {
		mpfr_set (y, u.y, MPFR_RNDD);
		mpfr_set (z, u.z, MPFR_RNDU);
	    }
	accuracy = u.accuracy < T*LOG  ? u.accuracy : T*LOG ; 
    }
    else {
	// precision of u < precision of this
	// copy of the fields
	mpfr_set (x, u.x, MPFR_RNDN);
	mpfr_set (y, u.y, MPFR_RNDN);
	mpfr_set (z, u.z, MPFR_RNDN);
	
	accuracy = u.accuracy;
	
	if (mpfr_equal_p(u.x, u.y)!=0 &&  mpfr_equal_p(u.x, u.z)!=0) {
	    // the 3 fields are equal
	    // perturbation of the fields to  avoid an increase of accuracy
	    // the perturbation is on the bit S+1.
	    // the perturbation doesn't change the accuracy value of the stochastic number
	    // if the three fields are different, the accuracy value is embedded in the stochastic
	    // number and no perturbation is required

	    perturbationLastBit(this, u);

	}
    }
    return *this;
} 

mp_st<T>& operator=(const mp_st<T> &u) // copy
{
    if ( this != &u ) {
	mpfr_set(x, u.x, MPFR_RNDN);
	mpfr_set(y, u.y, MPFR_RNDN);
	mpfr_set(z, u.z, MPFR_RNDN);
	accuracy=u.accuracy;
    }

    return *this;
}

mp_st<T>& operator=(mp_st<T>&& u) //move
{
    if ( this != &u ){
        mpfr_swap( x, u.x );
        mpfr_swap( y, u.y );
        mpfr_swap( z, u.z );
        std::swap( accuracy, u.accuracy );
    }
    return *this;
}


//----------------------------------------------------------
template <typename argT>
mp_st<T>& operator=(const argT &u) // copy
{

    mpfr_prec_t p = SAM::mp_st_nbbit<argT>::prec(u);
    if ( p >=  (mpfr_prec_t) T ) {
	
      p = T; 
      if (SAM_RANDOM) SAM::set_mpfr (x, u, MPFR_RNDU);
      else SAM::set_mpfr (x, u, MPFR_RNDD);
      if ( SAM_RANDOM )
	{
	    SAM::set_mpfr (y, u, MPFR_RNDU);
	    SAM::set_mpfr (z, u, MPFR_RNDD);
	}
      else
	{
	    SAM::set_mpfr (y, u, MPFR_RNDD);
	    SAM::set_mpfr (z, u, MPFR_RNDU);
	}
    }
   else {
       // p < T
       // the size of mp_st<T>  is greater than the size of u
       // u can be exactly coded in the mp_st<T>
       SAM::set_mpfr (x, u, MPFR_RNDN);
       SAM::set_mpfr (y, u, MPFR_RNDN);
       SAM::set_mpfr (z, u, MPFR_RNDN);

       // the 3 fields are equal
       // if u is an integer, the stochastic number is exact
       // if u is a floating point number (float, double, mpfr_t) a  perturbation of the fields to  avoid an increase
       // of the accuracy is necessary
       // the perturbation is on the bit S+1.
       // the perturbation doesn't change the accuracy value of the stochastic number

       p=perturbationLastBit(this, u);
   }
       
    accuracy = (int)( p * LOG );
    return *this;
}


////////////////


